/*  RemoteAlgManager.cpp -- basic interface class between software agent and hardware token 
 
    This is part of SecureFW, protection interface for which allows controled usage of secured code.
    See http://www.svenda.com/petr/securefw.html for more information
    Copyright (C) 2004  Petr Svenda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please, report any bugs to author <petr@svenda.com>
/**/
 
#include "RemoteAlgManager.h"
#include "SecFunctProvider.h"
#include "md5.h"

CRemoteAlgManager::CRemoteAlgManager() {
    m_actualSWAProfile.clear();
    m_actualRemoteApplet.clear();

    memcpy(m_aesCipherAuth.m_IV, IV, 16); 
    memcpy(m_aesCipherTrans.m_IV, IV, 16); 
}

CRemoteAlgManager::~CRemoteAlgManager() {
    int         status = STAT_OK;

    // DISCONNECT FROM TOKEN EVERYTIME
    if ((status = m_apduManager.CloseSession()) == STAT_OK) {
        // OK
    }
}

int CRemoteAlgManager::Connect(REMOTE_APPLET* pRemoteApplet) {
    // 0. DISCONNECT PREVIUS CONNECTION
    // 1. CONNECT TO REQUIRED TOKEN 
    // 2. SELECT REQUIRED APPLET

    int         status = STAT_OK;
    CARDAPDU    apdu;
        
    if (m_actualRemoteApplet.valid) {
        // DISCONNECT PREVIOUS CONNECTION
        if ((status = Disconnect()) == STAT_OK) {
            status = Connect(pRemoteApplet);
        }
    }
    else {
        // CONNECT TO REQUIRED TOKEN 
        if ((status = m_apduManager.OpenSession(pRemoteApplet->slotID, pRemoteApplet->acceptedCards)) == STAT_OK) {
            // OK
        }
	        
        // SELECT REQUIRED APPLET
        if (status == STAT_OK) {
	        apdu.cla = 0x00;
	        apdu.ins = 0xA4;
	        apdu.p1 = 0x04;
	        apdu.p2 = 0x00;
	        apdu.le = 0x00;
	        apdu.lc = pRemoteApplet->appletAIDLen;
	        memcpy(apdu.DataIn, pRemoteApplet->appletAID, pRemoteApplet->appletAIDLen);
        
            if ((status = m_apduManager.ExchangeAPDU(&apdu)) == STAT_OK) {
                // OK
                m_actualRemoteApplet = *pRemoteApplet;
                m_actualRemoteApplet.valid = TRUE;
            } 
        }
    }

    return status;
}

int CRemoteAlgManager::Disconnect() {
    int         status = STAT_OK;

    // DISCONNECT FROM TOKEN
    if ((status = m_apduManager.CloseSession()) == STAT_OK) {
        // OK
        m_actualRemoteApplet.clear();
        m_actualSWAProfile.clear();
    }

    return status;
}

int CRemoteAlgManager::ProcessAuth(SWA_PROFILE* pSWAProfile, BYTE debugFlags) {
    //
    //  PROCESS AUTHENTIZATION BETWEEN HARDWARE TOKEN AND SOFTWARE AGENT AS DEFINED IN SEAUT PROTOCOL
    //
    int         status = STAT_OK;
    CARDAPDU    apdu;
        
    if (m_actualRemoteApplet.valid) {
        BYTE        randomNSWA[RANDOM_NONCE_LENGTH];
        BYTE        randomNSWA_message[RANDOM_NONCE_LENGTH];
        BYTE        randomNHWT[RANDOM_NONCE_LENGTH];
        BYTE        swaUUID_message[SWA_UUID_LENGTH];
        BYTE        offset;
        BYTE        encryptOffset;
        DWORD       dataLen;
        DWORD       dataLenMax;
        int         i;
        MD5_CTX		mdContext;
    
        srand((unsigned int) time(NULL));
        // GENERATE RANDOM NOUNCE
        for (i = 0; i < RANDOM_NONCE_LENGTH; i++) randomNSWA[i] = rand() % 256;
        for (i = 0; i < RANDOM_NONCE_LENGTH; i++) randomNHWT[i] = rand() % 256;

        //
        // SEND 1. AUTH MESSAGE {UUIDHWT, UUIDSWA, NSWA}
        //
        if (status == STAT_OK) {
	        apdu.cla = CLA_CARD_DEFAULT;
	        apdu.ins = INS_CARD_SEAUT;
	        apdu.p1 = 0x01;
            apdu.p2 = debugFlags;
            
            apdu.le = 0;
            memset(apdu.DataOut, 0, sizeof(apdu.DataOut));
        
            // INPUT DATA           
            offset = 0;
            memcpy(apdu.DataIn + offset, m_actualRemoteApplet.hwtUUID, m_actualRemoteApplet.hwtUUIDLen);
            offset += m_actualRemoteApplet.hwtUUIDLen;
            memcpy(apdu.DataIn + offset,  pSWAProfile->swaUUID,  pSWAProfile->swaUUIDLen);
            offset += pSWAProfile->swaUUIDLen;
            memcpy(apdu.DataIn + offset, randomNSWA, RANDOM_NONCE_LENGTH);
            offset += RANDOM_NONCE_LENGTH;
            apdu.lc = offset;

            if ((status = m_apduManager.ExchangeAPDU(&apdu)) == STAT_OK) {
                // OK
            } 
            else _ASSERT(FALSE);
        }
    
        //
        // RECEIVE 2. AUTH MESSAGE {UUIDHWT, UUIDSWA, decryptKeySWA(NHWT, NSWA, UUIDSWA)}
        //
        if (status == STAT_OK) {
            offset = 0;

            // VERIFY CORRECTNES OF UUIDHWT
            if (status == STAT_OK) {
                if (memcmp(m_actualRemoteApplet.hwtUUID, apdu.DataOut, m_actualRemoteApplet.hwtUUIDLen) != 0) {
                    status = STAT_INCORRECT_HWT;
                    _ASSERT(FALSE);
                }
                offset += m_actualRemoteApplet.hwtUUIDLen; 
            }
        
            // VERIFY CORRECTNES OF UUIDSWA
            if (status == STAT_OK) {
                if (memcmp(pSWAProfile->swaUUID, apdu.DataOut + offset,  pSWAProfile->swaUUIDLen) != 0) {
                    status = STAT_INCORRECT_SWA;
                    _ASSERT(FALSE);
                }
                offset += pSWAProfile->swaUUIDLen; 
            }
        

            // DECRYPT USING WBACR AES DECRYPT KEY {authDecryptKeySWA(NHWT, NSWA, UUIDSWA)}
            if (status == STAT_OK) {
                dataLen = apdu.le - offset;
                if (debugFlags & FLAG_DEBUG_DISABLE_IOBUFFER_SECURING) {
                    // DEBUG MODE, NO BUFFER SECURING
                }
                else {
                    // NO PADDING USED !!!
                    if ((status = m_aesCipherAuth.WBACRAES128_DecryptCBC(apdu.DataOut + offset, &dataLen, FALSE)) != STAT_OK) {
                        _ASSERT(FALSE);
                    }
                }
            }

            // CHECK EXPECTED LENGTH OF DECRYPTED PART
            if (status == STAT_OK) {    
                if (dataLen != (DWORD) (RANDOM_NONCE_LENGTH + RANDOM_NONCE_LENGTH + pSWAProfile->swaUUIDLen)) {
                    // INCORECT RESPONSE LENGTH
                    status = STAT_DATA_INCORRECT_LENGTH;
                    _ASSERT(FALSE);
                }
            }

            // GET NHWT RANDOM NOUNCE
            if (status == STAT_OK) {
                memcpy(randomNHWT, apdu.DataOut + offset, RANDOM_NONCE_LENGTH); 
                offset += RANDOM_NONCE_LENGTH;
            }

            // COMPARE FRESHNESS OF NSWA
            if (status == STAT_OK) {
                memcpy(randomNSWA_message, apdu.DataOut + offset, RANDOM_NONCE_LENGTH); 
                if (memcmp(apdu.DataOut + offset, randomNSWA, RANDOM_NONCE_LENGTH) != 0) {
                    _ASSERT(FALSE);
                    status = STAT_SEAUT_FRESHNESS_FAIL;
                }
                offset += RANDOM_NONCE_LENGTH;
            }

            // COMPARE CORRECTNESS OF UUIDSWA
            if (status == STAT_OK) {
                memcpy(swaUUID_message, apdu.DataOut + offset, pSWAProfile->swaUUIDLen); 
                if (memcmp(apdu.DataOut + offset, pSWAProfile->swaUUID, pSWAProfile->swaUUIDLen) != 0) {
                    status = STAT_INCORRECT_SWA;
                }
                offset += pSWAProfile->swaUUIDLen;
            }

            // CONSTRUCT SESSION KEY AS Kr = HASH(NHWT | NSWA | UUIDSWA | NSWA' | UUIDSWA')
            if (status == STAT_OK) {
                BYTE*    sessionKeyMaterial = NULL;
                DWORD    sessionKeyMaterialLen = 3 * RANDOM_NONCE_LENGTH + 2 * pSWAProfile->swaUUIDLen;
            
                sessionKeyMaterial = new BYTE[sessionKeyMaterialLen];
                memset(sessionKeyMaterial, 0, sessionKeyMaterialLen);
            
                offset = 0;
                memcpy(sessionKeyMaterial + offset, randomNHWT, RANDOM_NONCE_LENGTH);
                offset += RANDOM_NONCE_LENGTH;
                memcpy(sessionKeyMaterial + offset, randomNSWA_message, RANDOM_NONCE_LENGTH);
                offset += RANDOM_NONCE_LENGTH;
                memcpy(sessionKeyMaterial + offset, swaUUID_message, pSWAProfile->swaUUIDLen);
                offset += pSWAProfile->swaUUIDLen;
                memcpy(sessionKeyMaterial + offset, randomNSWA, RANDOM_NONCE_LENGTH);
                offset += RANDOM_NONCE_LENGTH;
                memcpy(sessionKeyMaterial + offset, pSWAProfile->swaUUID, pSWAProfile->swaUUIDLen);
                offset += pSWAProfile->swaUUIDLen;
            
            
                // CREATE SESSION KEY USING MD5 HASH 
                _ASSERT(MD5_DIGEST_LENGTH == SESSION_KEY_LEN);
	            MD5Init(&mdContext);
	            MD5Update(&mdContext, sessionKeyMaterial, offset);
	            MD5Final(&mdContext);
	            memcpy(m_sessionKey, mdContext.digest, MD5_DIGEST_LENGTH); 

                if (sessionKeyMaterial) {
                    memset(sessionKeyMaterial, 0, sessionKeyMaterialLen);
                    delete[] sessionKeyMaterial;
                }
            }
        }

        //
        // SEND 3. AUTH MESSAGE {UUIDHWT, UUIDSWA, authEncryptKeySWA(NSWA, NHWT))}
        //
        if (status == STAT_OK) {
	        apdu.cla = CLA_CARD_DEFAULT;
	        apdu.ins = INS_CARD_SEAUT;
	        apdu.p1 = 0x03;
            apdu.p2 = debugFlags;
        
            offset = 0;
            // UUIDHWT
            memcpy(apdu.DataIn + offset, m_actualRemoteApplet.hwtUUID, m_actualRemoteApplet.hwtUUIDLen);
            offset += m_actualRemoteApplet.hwtUUIDLen;
            // UUIDSWA
            memcpy(apdu.DataIn + offset, pSWAProfile->swaUUID, pSWAProfile->swaUUIDLen);
            offset += pSWAProfile->swaUUIDLen;
            encryptOffset = offset;

            // NSWA
            memcpy(apdu.DataIn + offset, randomNSWA, RANDOM_NONCE_LENGTH);
            offset += RANDOM_NONCE_LENGTH;
            // NHWT
            memcpy(apdu.DataIn + offset, randomNHWT, RANDOM_NONCE_LENGTH);
            offset += RANDOM_NONCE_LENGTH;

            // ENCRYPT authEncryptKeySWA(NSWA, NHWT)
            if (status == STAT_OK) {
                dataLen = 2 * RANDOM_NONCE_LENGTH;
                dataLenMax = APDU_MAX_INPUT_DATA_LEN;
                if (debugFlags & FLAG_DEBUG_DISABLE_IOBUFFER_SECURING) {
                    // DEBUG MODE, NO BUFFER SECURING
                }
                else {
                    // NO PADDING USED !!!
                    if ((status = m_aesCipherAuth.WBACRAES128_EncryptCBC(apdu.DataIn + encryptOffset, &dataLen, &dataLenMax, FALSE)) != STAT_OK) {
                        _ASSERT(FALSE);
                    }
                    offset = (BYTE) (dataLen + m_actualRemoteApplet.hwtUUIDLen + pSWAProfile->swaUUIDLen);
                }
            }

            if (status == STAT_OK) {
                apdu.lc = offset;
                apdu.le = 0;
        
                if ((status = m_apduManager.ExchangeAPDU(&apdu)) == STAT_OK) {
                    // OK
                } 
                else _ASSERT(FALSE);
            }
        }
    }
    else status = STAT_NO_CONNECTION;

    if (status == STAT_OK) {
        m_actualSWAProfile = *pSWAProfile;
        m_actualSWAProfile.valid = TRUE; 
    }

    return status;
}

int CRemoteAlgManager::StopAuth(BYTE debugFlags) {
    //
    //  STOP AUTHENTICATED COMMUNICATION
    //
    int         status = STAT_OK;
    DWORD       dataOffset;
    DWORD       dataLen;
    DWORD       dataLenMax;
    DWORD       cryptOffset;
    
       
    if (m_actualSWAProfile.valid) {
        CARDAPDU    apdu;

	    apdu.cla = CLA_CARD_DEFAULT;
	    apdu.ins = INS_CARD_SEAUT_STOP;
	    apdu.p1 = 0x00;
        apdu.p2 = debugFlags;
	    
        if ((m_actualSWAProfile.swaUUIDLen + m_actualRemoteApplet.hwtUUIDLen + MESSAGE_STOP_LENGTH) <= APDU_MAX_INPUT_DATA_LEN) { 
            //
            // ONE APDU
            //
            dataOffset = 0;
            // UUIDHWT    
            memcpy(apdu.DataIn + dataOffset, m_actualRemoteApplet.hwtUUID, m_actualRemoteApplet.hwtUUIDLen);
            dataOffset += m_actualRemoteApplet.hwtUUIDLen;    
            // UUIDSWA    
            memcpy(apdu.DataIn + dataOffset, m_actualSWAProfile.swaUUID, m_actualSWAProfile.swaUUIDLen);
            dataOffset += m_actualSWAProfile.swaUUIDLen;    
            cryptOffset = dataOffset;

            // "STOP" MESSAGE
            memcpy(apdu.DataIn + dataOffset, MESSAGE_STOP, MESSAGE_STOP_LENGTH);
            dataOffset += MESSAGE_STOP_LENGTH;    

            //
            //  SECURE PART OF apdu.DataIn OUTGOING BUFFER
            //
            dataLen = dataOffset - cryptOffset;
            if (debugFlags & FLAG_DEBUG_DISABLE_IOBUFFER_SECURING) {
                // DEBUG MODE, NO BUFFER SECURING
            }
            else {
                dataLenMax = APDU_MAX_INPUT_DATA_LEN;
                status = SecureRequest(apdu.DataIn + cryptOffset, &dataLen, &dataLenMax); 
            }
            dataOffset = (BYTE) (dataLen + cryptOffset);

            _ASSERT(dataOffset <= 0xFF);
            apdu.lc = (BYTE) (dataOffset & 0xFF);

            // CLEAR OUTPUT BUFFER
            memset(apdu.DataOut, 0, APDU_MAX_OUTPUT_DATA_LEN);
            apdu.le = 0x00;
            apdu.le = MESSAGE_STOP_LENGTH;
	    }
        else status = STAT_DATA_INCORRECT_LENGTH;
    
        if (status == STAT_OK) {
            status = m_apduManager.ExchangeAPDU(&apdu);
        }

        if (status == STAT_OK) {
            m_actualSWAProfile.clear();
        }
                
    }
    else status = STAT_NO_CONNECTION;

    return status;
}

int CRemoteAlgManager::ProcessMasterLicence(MASTER_LICENCE* pMasterLicence, BOOL outOfOrder, BYTE debugFlags) {
    //
    //  SEND MASTER LICENCE TO HARDWARE TOKEN
    //
    int         status = STAT_OK;

    CARDAPDU    apdu;
	apdu.cla = CLA_CARD_DEFAULT;
	apdu.ins = INS_CARD_PROCESS_MASTER_LICENCE;

    if (m_actualRemoteApplet.valid) {
        if ((pMasterLicence->dataLen + m_actualRemoteApplet.hwtUUIDLen) <= APDU_MAX_INPUT_DATA_LEN) { 
	        apdu.p1 = (outOfOrder) ? P1_LICENCE_OUTOFORDER : 0x00;
            apdu.p2 = debugFlags;
	    
            // UUIDHWT
            memcpy(apdu.DataIn, m_actualRemoteApplet.hwtUUID, m_actualRemoteApplet.hwtUUIDLen);
            // MASTER LICENCE DATA
            memcpy(apdu.DataIn + m_actualRemoteApplet.hwtUUIDLen, pMasterLicence->pData, pMasterLicence->dataLen);
    
            apdu.lc = (BYTE) (m_actualRemoteApplet.hwtUUIDLen + pMasterLicence->dataLen);
            apdu.le = 0x00;

            // SEND APDU
            if ((status = m_apduManager.ExchangeAPDU(&apdu)) == STAT_OK) {
                //OK
            }
	    }
        else status = STAT_LICENCE_TOO_LONG;
    }
    else status = STAT_NO_CONNECTION;
        
    return status;
}
 
int CRemoteAlgManager::ProcessAlgLicence(ALG_LICENCE* pAlgLicence, BOOL outOfOrder, BYTE debugFlags) {
    int         status = STAT_OK;

    CARDAPDU    apdu;
	apdu.cla = CLA_CARD_DEFAULT;
	apdu.ins = INS_CARD_PROCESS_ALG_LICENCE;

    if (m_actualRemoteApplet.valid) {
        if ((pAlgLicence->dataLen + m_actualRemoteApplet.hwtUUIDLen) <= APDU_MAX_INPUT_DATA_LEN) { 
	        apdu.p1 = outOfOrder ? P1_LICENCE_OUTOFORDER : 0x00;
            apdu.p2 = debugFlags;
	    
            // UUIDHWT
            memcpy(apdu.DataIn, m_actualRemoteApplet.hwtUUID, m_actualRemoteApplet.hwtUUIDLen);
            // ALG LICENCE DATA
            memcpy(apdu.DataIn + m_actualRemoteApplet.hwtUUIDLen, pAlgLicence->pData, pAlgLicence->dataLen);
    
            apdu.lc = (BYTE) (m_actualRemoteApplet.hwtUUIDLen + pAlgLicence->dataLen);
            apdu.le = 0x00;

            // SEND APDU
            if ((status = m_apduManager.ExchangeAPDU(&apdu)) == STAT_OK) {
                //OK
            }
	    }
        else status = STAT_LICENCE_TOO_LONG;
    }
    else status = STAT_NO_CONNECTION;
        
    return status;

}

int CRemoteAlgManager::DebugReset() {
    int         status = STAT_OK;

    if (m_actualRemoteApplet.valid) {
        CARDAPDU    apdu;

	    apdu.cla = CLA_CARD_DEFAULT;
	    apdu.ins = INS_CARD_DEBUG_RESET;
	    apdu.p1 = 0x00;
        apdu.p2 = 0x00;
	    apdu.lc = 0x00;
	    apdu.le = 0x00;

        // SEND APDU
        if ((status = m_apduManager.ExchangeAPDU(&apdu)) == STAT_OK) {
            //OK
        }
    }
    else status = STAT_NO_CONNECTION;


    return status;
}
 
int CRemoteAlgManager::ProcessAlgData(SHORT_DATA_BLOB* pAlgUUID, SHORT_DATA_BLOB* pInputData, SHORT_DATA_BLOB* pOutputData, BYTE debugFlags) {
    //
    //  SEND PROTECTED ALGORITHM INPUT DATA & RECEIVE RESPONSE,
    //  PREVIOUS AUTHENTIZATION IS REQUIRED 
    //
    int         status = STAT_OK;

    DWORD       dataOffset;
    DWORD       dataLen;
    DWORD       dataLenMax;
    DWORD       cryptOffset;

    if (m_actualSWAProfile.valid) {
        CARDAPDU    apdu;

	    apdu.cla = CLA_CARD_DEFAULT;
	    apdu.ins = INS_CARD_PROCESS_ALG_DATA;
	    apdu.p1 = 0x00;
        apdu.p2 = debugFlags;
	    apdu.lc = APDU_MAX_INPUT_DATA_LEN;

        if ((pAlgUUID->dataLen + m_actualSWAProfile.swaUUIDLen + m_actualRemoteApplet.hwtUUIDLen + pInputData->dataLen) <= APDU_MAX_INPUT_DATA_LEN) { 
            //
            // ONE APDU
            //
            dataOffset = 0;
            // UUIDHWT    
            memcpy(apdu.DataIn + dataOffset, m_actualRemoteApplet.hwtUUID, m_actualRemoteApplet.hwtUUIDLen);
            dataOffset += m_actualRemoteApplet.hwtUUIDLen;    
            // UUIDSWA    
            memcpy(apdu.DataIn + dataOffset, m_actualSWAProfile.swaUUID, m_actualSWAProfile.swaUUIDLen);
            dataOffset += m_actualSWAProfile.swaUUIDLen;    
            cryptOffset = dataOffset;

            // ALGORITHM UUID
            memcpy(apdu.DataIn + dataOffset, pAlgUUID->pData, pAlgUUID->dataLen);
            dataOffset += pAlgUUID->dataLen;    
            
            // INPUT DATA
            memcpy(apdu.DataIn + dataOffset, pInputData->pData, pInputData->dataLen);
            dataOffset += pInputData->dataLen;    

            //
            //  SECURE PART OF apdu.DataIn OUTGOING BUFFER
            //
            dataLen = dataOffset - cryptOffset;
            if (debugFlags & FLAG_DEBUG_DISABLE_IOBUFFER_SECURING) {
                // DEBUG MODE, NO BUFFER SECURING
            }
            else {
                dataLenMax = APDU_MAX_INPUT_DATA_LEN;
                status = SecureRequest(apdu.DataIn + cryptOffset, &dataLen, &dataLenMax); 
            }
            dataOffset = (BYTE) (dataLen + cryptOffset);

            _ASSERT(dataOffset <= 0xFF);
            apdu.lc = (BYTE) (dataOffset & 0xFF);

            // CLEAR OUTPUT BUFFER
            memset(apdu.DataOut, 0, APDU_MAX_OUTPUT_DATA_LEN);
            // NOTE: if dataLen > 255 overflow
            _ASSERT(dataLen <= 0xFF);
            apdu.le = (BYTE) pOutputData->dataLen;
	    }
        else {
            // MORE APDUS REQUIRED
            status = STAT_DATA_TOO_LONG;
        }
    
        if (status == STAT_OK) {
            if ((status = m_apduManager.ExchangeAPDU(&apdu)) == STAT_OK) {
                if (apdu.le >= (m_actualRemoteApplet.hwtUUIDLen + m_actualSWAProfile.swaUUIDLen)) {
                    // VERIFY CORRECTNES OF UUIDHWT
                    dataOffset = 0;
                    if (status == STAT_OK) {
                        if (memcmp(m_actualRemoteApplet.hwtUUID, apdu.DataOut + dataOffset, m_actualRemoteApplet.hwtUUIDLen) != 0) {
                            status = STAT_INCORRECT_HWT;
                            _ASSERT(FALSE);
                        }
                        dataOffset += m_actualRemoteApplet.hwtUUIDLen; 
                    }
        
                    // VERIFY CORRECTNES OF UUIDSWA
                    if (status == STAT_OK) {
                        if (memcmp(m_actualSWAProfile.swaUUID, apdu.DataOut + dataOffset,  m_actualSWAProfile.swaUUIDLen) != 0) {
                            status = STAT_INCORRECT_SWA;
                            _ASSERT(FALSE);
                        }
                        dataOffset += m_actualSWAProfile.swaUUIDLen; 
                    }

                    //
                    //  UNSECURE PART OF apdu.DataOut BUFFER {response}
                    //
                    dataLen = apdu.le - dataOffset;
                    if (debugFlags & FLAG_DEBUG_DISABLE_IOBUFFER_SECURING) {
                        // DEBUG MODE, NO BUFFER SECURING
                    }
                    else {
                        status = UnsecureResponse(apdu.DataOut + dataOffset, &dataLen); 
                    }
                }
                else status = STAT_RESPONSE_DATA_LENGTH_BAD;
            }
        }

        if (status == STAT_OK) {
            // COPY OUTPUT DATA
            if (dataLen > pOutputData->dataLen) {
                status = STAT_NOT_ENOUGHT_MEMORY;
            }
            else {
                memset(pOutputData->pData, 0, pOutputData->dataLen);
                memcpy(pOutputData->pData, apdu.DataOut + dataOffset, dataLen); 
            }
            pOutputData->dataLen = (USHORT) dataLen;
        }
    }  
    else {
        status = STAT_CHANNEL_NOT_EXISTS;
    }  

    return status;
}

//
//  PROTECTED
//

int CRemoteAlgManager::UpdateSessionKey(BYTE* pSessionKey) {
    int             status = STAT_OK;
	MD5_CTX		    mdContext;
    
    // ACTUALIZATION USING MD5 HASH 
    _ASSERT(MD5_DIGEST_LENGTH == SESSION_KEY_LEN);
	MD5Init (&mdContext);
	MD5Update (&mdContext, pSessionKey, SESSION_KEY_LEN);
	MD5Final (&mdContext);
	memcpy(pSessionKey, mdContext.digest, MD5_DIGEST_LENGTH); 

/*
    // MATYAS-MEYER-OSEAS HASH METHOD USING  BLOCK CIPHER
    // NOTE: ONLY ONE BLOCK IS HASHED => KEY IS EVERYTIME SAME (IV)
    // TODO: USE WBACR AES ENCRYPT PART AS BLOCK CIPHER FOR BETTER SECURITY   
    BYTE    tempSessionKey[SESSION_KEY_LEN];
    memcpy(tempSessionKey, pSessionKey, SESSION_KEY_LEN);
    if ((status = CSecFunctProvider::AES_EncryptOneBlock(tempSessionKey, (BYTE*) IV)) == STAT_OK) {
        for (int i = 0; i < SESSION_KEY_LEN; i++) pSessionKey[i] ^= tempSessionKey[i];
    }
/**/    

    return status;    
}

int CRemoteAlgManager::SecureRequest(BYTE* pRequest, DWORD* pbRequestLen, DWORD* pbRequestLenMax) {
    int     status = STAT_OK;
    DWORD   i = 0;    

    if (pRequest != NULL) {
        // SESSION KEY ACTUALIZATION
        UpdateSessionKey(m_sessionKey);

        // ENCRYPT USING WBACRAES KEY
        if (status == STAT_OK) {
            memcpy(m_aesCipherTrans.m_IV, m_sessionKey, BLOCK_LENGTH);
            if ((status = m_aesCipherTrans.WBACRAES128_EncryptCBC(pRequest, pbRequestLen, pbRequestLenMax)) != STAT_OK) {
                _ASSERT(FALSE);
            }
        }

        if (status == STAT_OK) {
            // REPEATED XOR WITH SESSION KEY
            for (i = 0; i < *pbRequestLen; i++) {
                pRequest[i] ^= m_sessionKey[i % SESSION_KEY_LEN];
            }
        }
    }         
    else status = STAT_DATA_CORRUPTED; 

    return status;        
}

int CRemoteAlgManager::UnsecureResponse(BYTE* pResponse, DWORD* pbResponseLen) {
    int     status = STAT_OK;
    DWORD   i = 0;
        
    if (pResponse != NULL) {
        // SESSION KEY ACTUALIZATION
        UpdateSessionKey(m_sessionKey);

        // REPEATED XOR WITH SESSION KEY
        for (i = 0; i < *pbResponseLen; i++) {
            pResponse[i] ^= m_sessionKey[i % SESSION_KEY_LEN];
        }

        memcpy(m_aesCipherTrans.m_IV, m_sessionKey, BLOCK_LENGTH);
        if ((status = m_aesCipherTrans.WBACRAES128_DecryptCBC(pResponse, pbResponseLen)) != STAT_OK) {
            _ASSERT(FALSE);
        }
    }
    else status = STAT_DATA_CORRUPTED;
          
    return status;        
}
